from kinepy.geometry import *


#  ---------------------------------------------------- P P P ----------------------------------------------------------

def p_p_p(system, cycle, rev1, rev2, rev3):
    sgn = system.signs[cycle]
    (s13, p13), (s11, p11) = rev1
    (s21, p21), (s22, p22) = rev2
    (s32, p32), (s33, p33) = rev3

    p11 = get_point(system, s11, p11)
    p13 = get_point(system, s13, p13)
    p21 = get_point(system, s21, p21)
    p22 = get_point(system, s22, p22)
    p32 = get_point(system, s32, p32)
    p33 = get_point(system, s33, p33)

    v12 = p21 - p11
    v32 = p22 - p32
    v13 = p33 - p13
    
    sq_a, sq_b, sq_c = sq_mag(v12), sq_mag(v13), sq_mag(v32)
    inv_a, inv_b, inv_c = sq_a ** -.5, sq_b ** -.5, sq_c ** -.5
    alpha = sgn * np.arccos(0.5 * (sq_a + sq_b - sq_c) * inv_a * inv_b)

    theta = get_angle2(v12, inv_a) + alpha
    gamma = theta - get_angle2(v13, inv_b)
    mat = rot(gamma)
    change_ref(system, s13, gamma, mat, p13, p11)

    p33 = p11 + unit(theta) / inv_b
    gamma = get_angle(p21 - p33) - get_angle2(v32, inv_c)
    mat = rot(gamma)
    change_ref(system, s22, gamma, mat, p32, p33)
    
    for p in cycle:
        p = system.joints[p]
        p.angle = system.get_ref(p.s2) - system.get_ref(p.s1)
        make_continuous(p.angle)
    
    return sum((system.eqs[s] for s in (s11, s22, s33)), ())
    

#  ---------------------------------------------------- G P P ----------------------------------------------------------

def g_p_p(system, cycle, pri, rev2, rev3):
    (s1p, alpha1p, d1p), (s1, alpha1, d1) = pri
    (s21, p21), (s22, p22) = rev2
    (s32, p32), (s33, p33) = rev3

    sgn = system.signs[cycle] * (2 * (s1 == system.joints[cycle[0]].s1) - 1)
    
    gamma = system.get_ref(s1) + alpha1 - system.get_ref(s1p) - alpha1p
    change_ref2(system, s1p, gamma, rot(gamma), get_point(system, s33, p33))

    p21, p22 = get_point(system, s21, p21), get_point(system, s22, p22)
    v = p21 - system.get_origin(s1) - (d1 - d1p) * z_cross(unit(system.get_ref(s1) + alpha1)) + system.get_origin(s1p)
    # v1 + rot(...) . v3

    v2 = p22 - get_point(system, s32, p32)
    mat_mul_r(rot(-system.get_ref(s1) - alpha1), v)  # v <- (X0; y)
    
    sq_z = sq_mag(v2)
    inv_z = sq_z ** -.5
    dx = sgn * (sq_z - v[1] ** 2) ** .5
    
    theta2 = np.arccos(-sgn * (1 - (v[1] * inv_z) ** 2) ** .5) * (2 * (v[1] >= 0) - 1)
    theta3 = get_angle2(v2, inv_z)
    
    gamma = theta2 + alpha1 + system.get_ref(s1) - theta3
    change_ref(system, s22, gamma, rot(gamma), p22, p21)
    trans(system, s1p, get_point(system, s32, p32))

    for p in cycle[1:]:
        p = system.joints[p]
        p.angle = system.get_ref(p.s2) - system.get_ref(p.s1)
        make_continuous(p.angle)
    pri = system.joints[cycle[0]]
    pri.delta = (v[0, :] + dx) * (2 * (s1 == pri.s1) - 1)
    
    return sum((system.eqs[s] for s in (s1, s22, s33)), ())


#  ---------------------------------------------------- P P G ----------------------------------------------------------

def p_p_g(system, cycle, rev1, rev2, pri):
    (s13, p13), (s11, p11) = rev1
    (s21, p21), (s22, p22) = rev2
    (s1, alpha1, d1), (s1p, alpha1p, d1p) = pri

    sgn = system.signs[cycle] * (2 * (s1 == system.joints[cycle[2]].s1) - 1)
    
    v1 = get_point(system, s21, p21) - get_point(system, s11, p11)
    v2 = get_point(system, s22, p22) - system.get_origin(s1) + \
        (d1p - d1) * unit(np.pi * .5 + system.get_ref(s1) + alpha1 - system.get_ref(s1p) - alpha1p)
    v = system.get_origin(s1p) - get_point(system, s13, p13)  # = v3
    
    mat_mul_r(rot(system.get_ref(s1) + alpha1 - system.get_ref(s1p) - alpha1p), v)
    v += v2
    
    mat_mul_r(rot(-system.get_ref(s1) - alpha1), v)  # v <- (X0; y)
    sq_z = sq_mag(v1)
    inv_z = sq_z ** -.5
    dx = sgn * (sq_z - v[1] ** 2) ** .5
    theta2 = (2 * (v[1] > 0) - 1) * np.arccos(-sgn * (1 - (v[1] * inv_z) ** 2) ** .5)
    theta3 = get_angle2(v1, inv_z)
    
    gamma = theta3 - theta2 - system.get_ref(s1) - alpha1
    change_ref(system, s22, gamma, rot(gamma), get_point(system, s22, p22), get_point(system, s21, p21))
    gamma = theta3 - theta2 - system.get_ref(s1p) - alpha1p
    change_ref(system, s13, gamma, rot(gamma), get_point(system, s13, p13), get_point(system, s11, p11))
    
    for p in cycle[:2]:
        p = system.joints[p]
        p.angle = system.get_ref(p.s2) - system.get_ref(p.s1)
    pri = system.joints[cycle[2]]
    pri.delta = (v[0] + dx) * (2 * (s1 == pri.s1) - 1)
    
    return sum((system.eqs[s] for s in (s11, s22, s1p)), ())


#  ---------------------------------------------------- P G G ----------------------------------------------------------
    
def p_g_g(system, cycle, rev, pri1, pri2):
    (s13, p13), (s11, p11) = rev
    (s1, alpha1, d1), (s1p, alpha1p, d1p) = pri1
    (s2, alpha2, d2), (s2p, alpha2p, d2p) = pri2
    
    gamma2 = (gamma1 := system.get_ref(s1) + alpha1 - system.get_ref(s1p) - alpha1p) + \
        system.get_ref(s2) + alpha2 - system.get_ref(s2p) - alpha2p
    
    change_ref2(system, s1p, gamma1, rot(gamma1), system.get_origin(s1p))
    change_ref(system, s13, gamma2, rot(gamma2), get_point(system, s13, p13), get_point(system, s11, p11))
    
    ux, uy = unit(system.get_ref(s1) + alpha1), unit(system.get_ref(s2) + alpha2)
    v1 = get_point(system, s11, p11) - system.get_origin(s1) - (d1 - d1p) * z_cross(ux)
    v2 = system.get_origin(s1p) - system.get_origin(s1p) - (d2 - d2p) * z_cross(uy)
    v3 = system.get_origin(s2p) - get_point(system, s13, p13)
    
    x, y = mat_mul_n(inv_mat(ux, uy), v1 + v2 + v3)
    
    trans(system, s1p, system.get_origin(s1) + (d1 - d1p) * z_cross(ux) + x * ux)
    
    g1 = system.joints[cycle[1]]
    g1.delta = x * (2 * (s1 == g1.s1) - 1)
    g2 = system.joints[cycle[2]]
    g2.delta = y * (2 * (s2 == g2.s1) - 1)
    p = system.joints[cycle[0]]
    p.angle = system.get_ref(p.s2) - system.get_ref(p.s1)
    
    return sum((system.eqs[s] for s in (s11, s1p, s2p)), ())


#  ---------------------------------------------------- G G P ----------------------------------------------------------

def g_g_p(system, cycle, pri1, pri2, rev):
    (s1p, alpha1p, d1p), (s1, alpha1, d1) = pri1
    (s2, alpha2, d2), (s2p, alpha2p, d2p) = pri2
    (s12, p12), (s13, p13) = rev
    
    gamma1, gamma2 = (
        system.get_ref(s1) + alpha1 - system.get_ref(s1p) - alpha1p,
        system.get_ref(s2) + alpha2 - system.get_ref(s2p) - alpha2p
    )
    
    change_ref2(system, s2p, gamma2, rot(gamma2), get_point(system, s12, p12))
    change_ref2(system, s1p, gamma1, rot(gamma1), system.get_ref(s1p))
    
    ux, uy = unit(system.get_ref(s1) + alpha1), unit(system.get_ref(s2) + alpha2)
    
    v1 = system.get_origin(s2) + (d2 - d2p) * z_cross(uy) - system.get_origin(s1) - (d1 - d1p) * z_cross(ux)
    v2 = -system.get_origin(s2p)
    v3 = -get_point(system, s13, p13)
    
    x, y = mat_mul_n(inv_mat(ux, uy), v1 + v2 + v3)
    
    trans(system, s1p, system.get_origin(s1) + (d1 - d1p) * z_cross(ux) + x * ux)
    trans(system, s2p, get_point(system, s13, p13))
    
    g1 = system.joints[cycle[0]]
    g1.delta = x * (2 * (s1 == g1.s1) - 1)
    g2 = system.joints[cycle[1]]
    g2.delta = y * (2 * (s2 == g2.s2) - 1)
    p = system.joints[cycle[2]]
    p.angle = system.get_ref(p.s2) - system.get_ref(p.s1)
    make_continuous(p.angle)
    
    return sum((system.eqs[s] for s in (s1, s1p, s2p)), ())
   

#  ------------------------------------------------ SP G ---------------------------------------------------------------

def sp_g_1(system, cycle, pin, pri):
    (s12, p12), (s, alpha, d) = pin
    (s1, alpha1, d1), (s1p, alpha1p, d1p) = pri
    
    gamma = system.get_ref(s1) + alpha1 - system.get_ref(s1p) - alpha1p
    change_ref2(system, s1p, gamma, rot(gamma), system.get_origin(s1p))
    
    ux, uy = unit(alpha + system.get_ref(s)), unit(alpha1 + system.get_ref(s1))
    
    v1 = system.get_origin(s) + d * z_cross(ux) - system.get_origin(s1) - (d1 - d1p) * z_cross(uy)
    v2 = -get_point(system, s12, p12)
    
    x, y = mat_mul_n(inv_mat(ux, uy), v1 + v2)
    trans(system, s1p, system.get_origin(s1) + (d1 - d1p) * z_cross(uy) + y * uy)
    
    sp = system.joints[cycle[0]]
    sp.delta = -x
    sp.angle = system.get_ref(sp.s2) - system.get_ref(sp.s1)
    make_continuous(sp.angle)
    g = system.joints[cycle[1]]
    g.delta = y * (2 * (s1 == g.s1) - 1)
    
    return system.eqs[s] + system.eqs[s1p]


def sp_g_2(system, cycle, pin, pri):
    (s, alpha, d), (s11, p11) = pin
    (s1, alpha1, d1), (s1p, alpha1p, d1p) = pri
    
    gamma = system.get_ref(s1) + alpha1 - system.get_ref(s1p) - alpha1p
    change_ref2(system, s1p, gamma, rot(gamma), system.get_origin(s1p))
    
    ux, uy = unit(alpha + system.get_ref(s)), unit(alpha1 + system.get_ref(s1))
    
    v1 = system.get_origin(s1) + (d1 - d1p) * z_cross(uy) - get_point(system, s11, p11)
    v2 = system.get_origin(s) + d * ux
    
    x, y = mat_mul_n(inv_mat(ux, uy), v1 + v2)
    trans(system, s1p, system.get_origin(s1) + (d1 - d1p) * z_cross(uy) - y * uy)
    
    sp = system.joints[cycle[0]]
    sp.delta = -x
    sp.angle = system.get_ref(sp.s2) - system.get_ref(sp.s1)
    make_continuous(sp.angle)
    g = system.joints[cycle[1]]
    g.delta = y * (-2 * (s1 == g.s1) + 1)
    
    return system.eqs[s] + system.eqs[s1]


def sp_g(system, cycle, pin, pri):
    return (sp_g_1 if len(pin[0]) == 2 else sp_g_2)(system, cycle, pin, pri)


#  ---------------------------------------------------- T P ------------------------------------------------------------

def t_p(system, cycle, rec, rev):
    (s2, _), (s1, alpha) = rec
    (s11, p11), (s12, p12) = rev
    
    gamma = system.get_ref(s1) + alpha - system.get_ref(s2)
    change_ref(system, s12, gamma, rot(gamma), get_point(system, s12, p12), get_point(system, s11, p11))
    
    p = system.joints[cycle[1]]
    p.angle = system.get_ref(p.s2) - system.get_ref(p.s1)
    t = system.joints[cycle[0]]
    v, theta = system.get_origin(t.s2) - system.get_origin(t.s1), system.get_ref(t.s1)
    
    mat_mul_r(inv_mat(unit(t.base[0] + theta), unit(t.base[1] + theta)), v)
    t.delta = v[:, 0, :]
    
    return system.eqs[s1] + system.eqs[s2]
    

#  -------------------------------------------------- SP P -------------------------------------------------------------

def sp_p_1(system, cycle, pin, rev):
    sgn = system.signs[cycle]
    
    (s, alpha, d), (s11, p11) = pin
    (s21, p21), (s22, p22) = rev
    
    v1 = get_point(system, s21, p21) - get_point(system, s11, p11)
    v2 = get_point(system, s22, p22) - system.get_origin(s) - d * unit(alpha + system.get_ref(s) + np.pi * .5)
    mat_mul_r(rot(-alpha - system.get_ref(s)), v2)   # v2 <- (X0; y)
    
    sq_z = sq_mag(v1)
    inv_z = sq_z ** -.5
    
    dx = sgn * (sq_z - v2[1] ** 2) ** .5
    theta2 = (2 * (v2[1] > 0) - 1) * np.arccos(-sgn * (1 - (v2[1] * inv_z) ** 2) ** .5)
    theta1 = get_angle2(v1, inv_z)
    
    gamma = theta1 - theta2 - alpha - system.get_ref(s)
    change_ref(system, s, gamma, rot(gamma), get_point(system, s22, p22), get_point(system, s21, p21))
    
    pin = system.joints[cycle[0]]
    pin.delta = v2[0] + dx
    pin.angle = system.get_ref(pin.s2) - system.get_ref(pin.s1)
    make_continuous(pin.angle)
    p = system.joints[cycle[1]]
    p.angle = system.get_ref(p.s2) - system.get_ref(p.s1)
    make_continuous(p.angle)
    
    return system.eqs[s] + system.eqs[s21]


def sp_p_2(system, cycle, pin, rev):
    sgn = system.signs[cycle]
    
    (s12, p12), (s, alpha, d) = pin
    (s21, p21), (s22, p22) = rev
    
    v1 = get_point(system, s21, p21) - system.get_origin(s) - d * unit(np.pi * .5 + alpha + system.get_ref(s))
    v2 = get_point(system, s22, p22) - get_point(system, s12, p12)
    
    mat_mul_r(rot(-alpha - system.get_ref(s)), v1)  # v1 <- (X0; y)
    sq_z = sq_mag(v2)
    inv_z = sq_z ** - .5
    
    dx = sgn * (sq_z - v1[1] ** 2) ** .5
    theta2 = (2 * (v1[1] > 0) - 1) * np.arccos(-sgn * (1 - (v1[1] * inv_z) ** 2) ** .5)
    theta1 = get_angle2(v2, inv_z)
    
    gamma = theta2 + alpha + system.get_ref(s) - theta1
    change_ref(system, s12, gamma, rot(gamma), get_point(system, s22, p22), get_point(system, s21, p21))
    
    pin = system.joints[cycle[0]]
    pin.delta = v1[0] + dx
    pin.angle = system.get_ref(pin.s2) - system.get_ref(pin.s1)
    make_continuous(pin.angle)
    p = system.joints[cycle[1]]
    p.angle = system.get_ref(p.s2) - system.get_ref(p.s1)
    make_continuous(p.angle)

    return system.eqs[s] + system.eqs[s22]


def sp_p(system, cycle, pin, rev):
    return (sp_p_1 if len(pin[0]) == 3 else sp_p_2)(system, cycle, pin, rev)


def pilot(system, index):
    return system.joints[index].pilot(system, system.indices[index])


kin = {
    'P_P_P': p_p_p,
    'P_P_G': p_p_g,
    'G_P_P': g_p_p,
    'G_G_P': g_g_p,
    'P_G_G': p_g_g,
    'SP_P': sp_p,
    'SP_G': sp_g,
    'T_P': t_p,
    'Pilot': pilot
}



from kinepy.linkage import RevoluteJoint, PrismaticJoint, PinSlotJoint, RectangularJoint

KINEMATICS, DYNAMICS, BOTH = 0, 1, 2
MAX_CYCLE = 3


class CompilationError(Exception):
    pass


def make_graph(system):
    graph: list[list[tuple[int, int]]] = [list() for _ in system.sols]
    for i, l_ in enumerate(system.joints):
        graph[l_.s1].append((l_.s2, i))
        graph[l_.s2].append((l_.s1, i))
    return graph


def set_distances(graph, eqs):
    d = [-1] * len(graph)
    d[0] = 0
    queue = [0]
    while queue:
        c = queue.pop(0)
        for s, _ in graph[c]:
            e = eqs[s]
            if d[e] == -1:
                d[e] = d[c] + 1
                queue.append(e)
    return d


def closest_piloted_joint(system, eqs, d, mode):
    d_min, index, eq_ = float('inf'), None, ()
    for l_i in (system.piloted, system.blocked, system.piloted)[mode]:
        l_ = system.joints[l_i]
        if (eq1 := eqs[l_.s1]) != (eq2 := eqs[l_.s2]) and min(d[eq1], d[eq2]) < d_min:
            d_min, index, eq_ = min(d[eq1], d[eq2]), l_i, ((eq1, l_.s1), (eq2, l_.s2))[::(2 * (d[eq1] < d[eq2]) - 1)]
            if d_min == 0:
                return d_min, index, eq_
    return d_min, index, eq_


def hash_cycle(cycle):  # maximum is 3-cycle
    return sum((MAX_CYCLE+1) ** l_.id_ for l_ in cycle)


valid_cycles = {
    hash_cycle((RevoluteJoint, RevoluteJoint, RevoluteJoint)),
    hash_cycle((RevoluteJoint, RevoluteJoint, PrismaticJoint)),
    hash_cycle((RevoluteJoint, PrismaticJoint, PrismaticJoint)),
    hash_cycle((PinSlotJoint, RevoluteJoint)),
    hash_cycle((PinSlotJoint, PrismaticJoint)),
    hash_cycle((RectangularJoint, RevoluteJoint))
}

signed_cycles = {
    hash_cycle((RevoluteJoint, RevoluteJoint, RevoluteJoint)),
    hash_cycle((RevoluteJoint, RevoluteJoint, PrismaticJoint)),
    hash_cycle((RevoluteJoint, PinSlotJoint))
}


def find_cycle(system, eqs, graph, d):
    queue = [0]
    closed = [False] * len(graph)
    while queue:
        c = queue.pop(0)
        nbs = {}
        for s, l_ in graph[c]:
            eq = eqs[s]
            if closed[eq]:
                continue
            # 2-cycle
            if eq in nbs and (h := hash_cycle((system.joints[nbs[eq]], system.joints[l_]))) in valid_cycles:
                return d[c], (nbs[eq], l_), (c, eq), h in signed_cycles
            if eq in nbs:
                raise CompilationError(f'Invalid Hyper-static 2-cycle encountered: {(nbs[eq], l_)}')
            nbs[eq] = l_
        # 3-cycle
        for nb, l1 in nbs.items():
            for s, l_ in graph[nb]:
                e = eqs[s]
                if closed[e]:
                    continue
                if e in nbs and (h := hash_cycle(tuple(system.joints[i] for i in (l1, nbs[e], l_)))) in valid_cycles:
                    return d[c], (l1, nbs[e], l_), (c, e, nb), h in signed_cycles
            queue.append(nb)
        closed[c] = True
    return float('inf'), None, (), False


def next_step(system, eqs, graph, d, mode):
    d1, i1, eq1 = closest_piloted_joint(system, eqs, d, mode)
    if not d1:
        # pilot that joint
        return i1, eq1, False
    d2, cycle, eq2, sgn = find_cycle(system, eqs, graph, d)
    if d1 < d2:
        # pilot the joint
        return i1, eq1, False
    # solve the cycle
    if d2 == float('inf'):
        graph_text = "\n".join(str(i) for i in graph)
        raise CompilationError(f'Nothing found in graph:\n{graph_text}\nEqs: {eqs}')
    return cycle, eq2, sgn


def align(cycle, cycle_eqs, cycle_indices, eqs):
    n_cycle = tuple(
        l_[::(2 * (eqs[l_[1][0]] == cycle_eqs[i]) - 1)] for i, l_ in enumerate((l_.identifier for l_ in cycle))
    )
    if len(cycle) == 2:
        if cycle[0].id_ < cycle[1].id_:
            return (n_cycle[1][::-1], n_cycle[0][::-1]), cycle_eqs, cycle_indices[::-1]
        return n_cycle, cycle_eqs, cycle_indices
    if cycle[0].id_ == cycle[1].id_ or cycle[1].id_ == cycle[2].id_:
        return n_cycle, cycle_eqs, cycle_indices
    n_cycle = n_cycle[1][::-1], n_cycle[0][::-1], n_cycle[2][::-1]
    cycle_eqs = cycle_eqs[0], cycle_eqs[2], cycle_eqs[1]
    cycle_indices = cycle_indices[1], cycle_indices[0], cycle_indices[2]
    return n_cycle, cycle_eqs, cycle_indices


def eq_union(eq, graph, eqs, _eqs):
    graph[eq[0]] = sum((tuple((s, l_) for s, l_ in graph[i] if eqs[s] not in eq) for i in eq), ())
    _eqs[eq[0]] = sum((_eqs[i] for i in eq), ())
    for i in sorted(eq[1:], reverse=True):
        for s in _eqs[i]:
            eqs[s] = eq[0]
        if i < len(graph) - 1:
            graph[i] = graph.pop()
            _eqs[i] = _eqs.pop()
            for s in _eqs[i]:
                eqs[s] = i
        else:
            graph.pop()
            _eqs.pop()


def compiler(system, mode=KINEMATICS):
    eqs, _eqs = list(range(len(system.sols))), list((i,) for i in range(len(system.sols)))
    graph = make_graph(system)
    kin_instr, dyn_instr = [], []
    d = set_distances(graph, eqs)
    final = [0] * len(eqs)
    while eqs != final:
        cycle, eq, signed = next_step(system, eqs, graph, d, mode)
        if isinstance(cycle, int):
            eq_, eq = tuple((_eqs[i], s) for i, s in eq), tuple(i for i, _ in eq)
            kin_instr.append(('Pilot', cycle))
            dyn_instr.insert(0, ('Block', cycle) + eq_)
        else:
            cycle, eq, cycle_indices = align(tuple(system.joints[i] for i in cycle), eq, cycle, eqs)
            eq_ = tuple(_eqs[i] for i in eq)
            tag = '_'.join(system.joints[l_].tag for l_ in cycle_indices)
            kin_instr.append((tag, cycle_indices,) + cycle)
            dyn_instr.insert(0, (tag, cycle_indices,) + cycle + eq_)

            if signed and cycle_indices not in system.signs:
                system.signs[cycle_indices] = 1
                print(f'Identified new signed cycle: {cycle_indices} ({tag}).\nChosen 1 as sign.')
        eq_union(eq, graph, eqs, _eqs)
        d = set_distances(graph, eqs)
    return (kin_instr, dyn_instr, (kin_instr, dyn_instr))[mode]



from kinepy.linkage import *
from kinepy.solid import *
from kinepy.compilation import compiler, DYNAMICS, BOTH
from kinepy.kinematic import kin
from kinepy.dynamic import Spring, AccelerationField, dyn
import json


class System:
    def __init__(self, sols=(), joints=(), piloted=(), blocked=(), signs=None, name=''):
        self.sols, self.joints = list(sols) if sols else [Solid(name='Ground')], list(joints)
        self.piloted, self.blocked = list(piloted), list(blocked)
        self.name = name if name else 'Main system'

        # Renumérotation des Sols
        for i, s in enumerate(self.sols):
            s.rep = i

        # Dictionnaires des noms
        self.named_sols = {s.name: i for i, s in enumerate(self.sols)}
        self.named_joints = {s.name: i for i, s in enumerate(joints)}

        # Référence dans les liaisons
        for l_ in self.joints:
            l_.system = self

        # Pilotage, mise à jour de l'entrée pour résolution cinématiue
        self.tot, self.indices = 0, {}
        for l_ in piloted:
            pil = []
            for _ in self.joints[l_].input_mode():
                pil.append(self.tot)
                self.tot += 1
            self.indices[l_] = tuple(pil)
        self.show_input()

        # Préparation cinématique
        self.input = None
        self.signs = dict() if signs is None else signs
        self.eqs = None
        self.kin_instr, self.dyn_instr = [], []

        # Préparation dynamique
        self.interactions = []

    def add_solid(self, j=0., m=0., g=(0., 0.), name=''):
        s = Solid(j, m, g, name, len(self.sols))
        self.named_sols[s.name] = len(self.sols)
        self.sols.append(s)
        return s
    
    def add_revolute(self, s1, s2, p1=(0., 0.), p2=(0., 0.)):
        if isinstance(s1, str):
            # Référence par le nom
            s1 = self.named_sols[s1]
    
        if isinstance(s2, str):
            # Référence par le nom
            s2 = self.named_sols[s2]
        
        p = RevoluteJoint(s1, s2, p1, p2)
        print(f'Added linkage {p}')
        self.named_joints[p.name] = len(self.joints)

        self.joints.append(p)
        p.system = self
        return p

    def add_prismatic(self, s1, s2, a1=0., d1=0., a2=0., d2=0.):
        if isinstance(s1, str):
            # Référence par le nom
            s1 = self.named_sols[s1]
        if isinstance(s2, str):
            # Référence par le nom
            s2 = self.named_sols[s2]
            
        g = PrismaticJoint(s1, s2, a1, d1, a2, d2)
        print(f'Added linkage {g}')
        self.named_joints[g.name] = len(self.joints)
        self.joints.append(g)
        g.system = self
        return g
    
    def add_pin_slot(self, s1, s2, a1=0., d1=0., p2=(0., 0.)):
        if isinstance(s1, str):
            # Référence par le nom
            s1 = self.named_sols[s1]
        if isinstance(s2, str):
            # Référence par le nom
            s2 = self.named_sols[s2]
            
        sp = PinSlotJoint(s1, s2, a1, d1, p2)
        print(f'Added linkage {sp}')
        self.named_joints[sp.name] = len(self.joints)

        self.joints.append(sp)
        sp.system = self
        return sp
    
    def add_rectangle(self, s1, s2, angle=0., base=(0., np.pi/2)):
        if isinstance(s1, str):
            # Référence par le nom
            s1 = self.named_sols[s1]
        if isinstance(s2, str):
            # Référence par le nom
            s2 = self.named_sols[s2]
        
        t = RectangularJoint(s1, s2, angle, base)
        print(f'Added linkage {t}')
        self.named_joints[t.name] = len(self.joints)
        self.joints.append(t)
        t.system = self
        return t

    def add_3dof(self, s1, s2):
        if isinstance(s1, str):
            # Référence par le nom
            s1 = self.named_sols[s1]
        if isinstance(s2, str):
            # Référence par le nom
            s2 = self.named_sols[s2]

        _3dof = ThreeDegreesOfFreedomJoint(s1, s2)
        print(f'Added linkage {_3dof}')
        self.named_joints[_3dof.name] = len(self.joints)
        self.joints.append(_3dof)
        _3dof.system = self
        return _3dof

    def add_joint(self, joint: Joint):
        self.named_joints[joint.name] = len(self.joints)
        self.joints.append(joint)
        joint.system = self
    
    def pilot(self, joint):
        if isinstance(joint, (tuple, list)):
            # Plusieurs liaisons en entrée
            for j in joint:
                if isinstance(j, Joint):
                    j = j.name
                if isinstance(j, str):
                    # Référence par le nom
                    j = self.named_joints[j]
                self.piloted.append(j)

                # Mse à jour de l'entrée pour résolution cinématiue
                pil = []
                for _ in self.joints[j].input_mode():
                    pil.append(self.tot)
                    self.tot += 1
                self.indices[j] = tuple(pil)

            return self.show_input()
        if isinstance(joint, Joint):
            joint = joint.name
        if isinstance(joint, str):
            # Référence par le nom
            joint = self.named_joints[joint]

        # Mse à jour de l'entrée pour résolution cinématiue
        pil = []
        for _ in self.joints[joint].input_mode():
            pil.append(self.tot)
            self.tot += 1
        self.indices[joint] = tuple(pil)

        self.piloted.append(joint)
        self.show_input()

    def block(self, joint):
        if isinstance(joint, (tuple, list)):
            for j in joint:
                if isinstance(j, Joint):
                    j = j.name
                if isinstance(j, str):
                    # Référence par le nom
                    j = self.named_joints[j]
                self.blocked.append(j)
        else:
            if isinstance(joint, Joint):
                joint = joint.name
            if isinstance(joint, str):
                # Référence par le nom
                joint = self.named_joints[joint]
            self.blocked.append(joint)

    def show_input(self):
        j = []
        for joint in self.piloted:
            for m in self.joints[joint].input_mode():
                j.append(m)
        print('Current input order:')
        print(f"({'; '.join(j)})")
    
    def reset(self, n):
        self.eqs = [(i,) for i in range(len(self.sols))]
        for sol in self.sols:
            sol.reset(n)
        for joint in self.joints:
            joint.reset(n)

    def get_origin(self, sol):
        return self.sols[sol].origin
    
    def get_ref(self, sol):
        return self.sols[sol].angle

    def compile(self):
        if not self.blocked or set(self.blocked) == set(self.piloted):
            self.kin_instr, self.dyn_instr = compiler(self, BOTH)
        else:
            self.kin_instr, self.dyn_instr = compiler(self), compiler(self, DYNAMICS)
        print('signs =', self.signs)
        
    def solve_kinematics(self, input_):
        if isinstance(input_, (list, tuple)):
            input_ = np.array(input_)
        if len(input_.shape) == 1:
            input_ = input_[np.newaxis, :]
        self.reset(input_.shape[1])
        self.input = input_
        for instr in self.kin_instr:
            eq = kin[instr[0]](self, *instr[1:])
            for s in eq:
                self.eqs[s] = eq

    def solve_dynamics(self, dt):
        for s in self.sols:
            og = s.get_point(s.g)
            s.mech_actions['Inertie'] = MechanicalAction(
                -s.m * derivative2_vec(og, dt), og, -s.j * derivative2(s.angle, dt)
            )
        for inter in self.interactions:
            inter.set_am(self)
        for instr in self.dyn_instr:
            dyn[instr[0]](self, *instr[1:])

    def add_acceleration_field(self, g, name='Gravity'):
        self.interactions.append(AccelerationField(g, name))

    def add_spring(self, k, l0, s1, s2, p1=(0, 0), p2=(2, 0)):
        if isinstance(s1, str):
            # Référence par le nom
            s1 = self.named_sols[s1]

        if isinstance(s2, str):
            # Référence par le nom
            s2 = self.named_sols[s2]

        self.interactions.append(Spring(k, l0, s1, s2, p1, p2))

    def get_data(self):
        return {
            'sols': [s.save() for s in self.sols],
            'joints': [l_.save() for l_ in self.joints],
            'piloted': self.piloted,
            'blocked': self.blocked,
            'signs': self.signs
        }

    def save(self, file):
        with open(file, 'w') as f:
            json.dump(self.get_data(), f)

    @classmethod
    def load(cls, file):
        with open(file) as f:
            data = json.load(f)
        data['sols'] = [Solid.load(s) for s in data['sols']]
        data['joints'] = [class_dict[name].load(d) for name, d in data['joints']]
        return cls(**data)

    def __repr__(self):
        return self.name



from kinepy.geometry import *
from kinepy.dynamic import trd, tmd, MechanicalAction


class Joint:
    id_ = -1

    def __init__(self, s1, s2, name):
        self.s1, self.s2, self.name = s1, s2, name
        self.system = None

    def __repr__(self):
        return self.name
    
    @classmethod
    def load(cls, data):
        return cls(**data)

    def get_data(self):
        pass

    def save(self):
        return self.__class__.__name__, self.get_data()

    def input_mode(self):
        return ()


class RevoluteJoint(Joint):
    id_ = 0
    tag = 'P'
    
    def __init__(self, s1, s2, p1, p2):
        Joint.__init__(self, s1, s2, f'Rev({s2}/{s1})')
        self.p1, self.p2 = p1, p2
        self.angle = None
    
    def input_mode(self):
        return f'{self.name}: Angle',
    
    @property
    def identifier(self):
        return (self.s1, self.p1), (self.s2, self.p2)

    def get_data(self):
        return {
            's1': self.s1,
            's2': self.s2,
            'p1': self.p1,
            'p2': self.p2
        }
    
    def reset(self, n):
        self.angle = np.zeros((n,), float)

    def pilot(self, system, index):
        self.angle = system.input[index[0]]
        theta = self.angle + system.get_ref(self.s1) - system.get_ref(self.s2)
        change_ref(system, self.s2, theta, rot(theta),
                   get_point(system, self.s2, self.p2), get_point(system, self.s1, self.p1))
        return system.eqs[self.s1] + system.eqs[self.s2]

    def block(self, system, eq1s1, eq2s2):
        (_, s1), (eq2, s2) = eq1s1, eq2s2
        p = self.point
        f = trd(system, eq2)
        m = tmd(system, p, eq2)
        system.sols[s1].mech_actions[self.name](MechanicalAction(f, p, m))
        system.sols[s2].mech_actions[self.name](MechanicalAction(-f, p, -m))

    @property
    def point(self):
        return get_point(self.system, self.s1, self.p1)


class PrismaticJoint(Joint):
    id_ = 1
    tag = 'G'

    def __init__(self, s1, s2, a1, d1, a2, d2):
        Joint.__init__(self, s1, s2, f'Pri({s2}/{s1})')
        self.a1, self.a2, self.d1, self.d2 = a1, a2, d1, d2
        self.delta = None
        
    def input_mode(self):
        return f'{self.name}: Length',
    
    @property
    def identifier(self):
        return (self.s1, self.a1, self.d1), (self.s2, self.a2, self.d2)

    def get_data(self):
        return {
            's1': self.s1,
            's2': self.s2,
            'a1': self.a1,
            'a2': self.a2,
            'd1': self.d1,
            'd2': self.d2,
            'name': self.name
        }
    
    def reset(self, n):
        self.delta = np.zeros((n,), float)

    def pilot(self, system, index):
        self.delta = system.input[index[0]]
        theta = (alpha := system.get_ref(self.s1) + self.a1) - system.get_ref(self.s2) - self.a2
        ux = unit(alpha)
        change_ref(system, self.s2, theta, rot(alpha), system.get_origin(self.s2),
                   system.get_origin(self.s1) + self.delta * ux + (self.d1 - self.d2) * z_cross(ux))
        return system.eqs[self.s1] + system.eqs[self.s2]

    def block(self, system, eq1s1, eq2s2):
        (_, s1), (eq2, s2) = eq1s1, eq2s2
        p = system.get_origin(self.s1) + self.d1 * unit(system.get_ref(self.s1) + self.a1 + np.pi * .5)
        f = trd(system, eq2)
        m = tmd(system, p, eq2)

        system.sols[s1].mech_actions[self.name](MechanicalAction(f, p, m))
        system.sols[s2].mech_actions[self.name](MechanicalAction(-f, p, -m))


class PinSlotJoint(Joint):
    id_ = 2
    tag = 'SP'

    def __init__(self, s1, s2, a1, d1, p2):
        Joint.__init__(self, s1, s2, f'Pin({s2}/{s1})')
        self.p2, self.a1, self.d1 = p2, a1, d1
        self. delta = self.angle = None
    
    def input_mode(self):
        return f'{self.name}: Length', f'{self.name}: Angle'
    
    @property
    def identifier(self):
        return (self.s2, self.p2), (self.s1, self.a1, self.d1)

    def get_data(self):
        return {
            's1': self.s1,
            'p2': self.p2,
            's2': self.s2,
            'a1': self.a1,
            'd1': self.d1
        }

    def reset(self, n):
        self.delta = np.zeros((n,))
        self.angle = np.zeros((n,))

    def pilot(self, system, index):
        self.delta, self.angle = system.input[index[0]], system.input[index[1]]
        theta = system.get_ref(self.s1) + self.angle - system.get_ref(self.s2)
        ux = unit(system.get_ref(self.s1) + self.a1)
        change_ref(system, self.s2, theta, rot(theta), get_point(system, self.s2, self.p2),
                   system.get_origin(self.s1) + self.delta * ux + self.d1 * z_cross(ux))
        return system.eqs[self.s1] + system.eqs[self.s2]

    @property
    def point(self):
        return get_point(self.system, self.s2, self.p2)

    def block(self, system, eq1s1, eq2s2):
        (_, s1), (eq2, s2) = eq1s1, eq2s2
        p = self.point
        f = trd(system, eq2)
        m = tmd(system, p, eq2)

        system.sols[s1].mech_actions[self.name](MechanicalAction(f, p, m))
        system.sols[s2].mech_actions[self.name](MechanicalAction(-f, p, -m))


class RectangularJoint(Joint):
    id_ = 3
    tag = 'T'

    def __init__(self, s1, s2, angle, base):
        Joint.__init__(self, s1, s2, f'Rect({s2}/{s1})')
        self.angle, self.base = angle, base
        self.delta = None

    def input_mode(self):
        return f'{self.name}: X', f'{self.name}: Y'
        
    @property
    def identifier(self):
        return (self.s1, self.angle), (self.s2, -self.angle)

    def get_data(self):
        return {
            's1': self.s1,
            's2': self.s2,
            'base': self.base,
            'angle': self.angle
        }
    
    def reset(self, n):
        self.delta = np.zeros((2, n), float)

    def pilot(self, system, index):
        self.delta = np.array((system.input[index[0]], system.input[index[1]]))
        theta = system.get_ref(self.s1) + self.angle - system.get_ref(self.s2)
        ux, uy = unit(system.get_ref(self.s1) + self.base[0]), unit(system.get_ref(self.s1) + self.base[1])
        change_ref(system, self.s1, theta, rot(theta), system.get_origin(self.s2),
                   system.get_origin(self.s1) + mat_mul_n(((ux[0], uy[0]), (ux[1], uy[1])), self.delta))
        return system.eqs[self.s1] + system.eqs[self.s2]

    def block(self, system, eq1s1, eq2s2):
        (_, s1), (eq2, s2) = eq1s1, eq2s2
        p = system.get_origin(self.s1)
        f = trd(system, eq2)
        m = tmd(system, p, eq2)

        system.sols[s1].mech_actions[self.name](MechanicalAction(f, p, m))
        system.sols[s2].mech_actions[self.name](MechanicalAction(-f, p, -m))


class ThreeDegreesOfFreedomJoint(Joint):
    id_ = 4
    tag = '3DoF'

    def __init__(self, s1, s2):
        Joint.__init__(self, s1, s2, f'3DoF({s2}/{s1})')
        self.delta, self.angle = None, None

    def get_data(self):
        return {'s1': self.s1, 's2': self.s2, 'name': self.name}

    def input_mode(self):
        return f'{self.name}: X', f'{self.name}: Y', f'{self.name}: Angle'

    @property
    def identifier(self):  # Ne devrait pas servir...
        return (self.s1,), (self.s2,)

    def reset(self, n):
        self.delta = np.zeros((2, n), float)
        self.angle = np.zeros((n,), float)

    def pilot(self, system, index):
        self.delta[0], self.delta[1] = system.input[index[0]], system.input[index[1]]
        self.angle = system.input[index[2]]
        theta = system.get_ref(self.s1) + self.angle - system.get_ref(self.s2)
        change_ref(system, self.s2, theta, rot(theta), system.get_origin(self.s2), system.get_origin(self.s1))
        return system.eqs[self.s1] + system.eqs[self.s2]

    def block(self, system, eq1s1, eq2s2):
        (_, s1), (eq2, s2) = eq1s1, eq2s2
        p = system.get_origin(self.s1)
        f = trd(system, eq2)
        m = tmd(system, p, eq2)

        system.sols[s1].mech_actions[self.name](MechanicalAction(f, p, m))
        system.sols[s2].mech_actions[self.name](MechanicalAction(-f, p, -m))


class_dict = {cls.__name__: cls for cls in Joint.__subclasses__()}



from kinepy.geometry import *


class MechanicalAction:
    def __init__(self, f: np.ndarray, p: np.ndarray, m):
        self.f, self.p, self.m = f, p, m

    def babar(self, point):
        return self.m + det(self.p - point, self.f)


class Interaction:
    name: str

    def set_am(self, system):
        pass


class AccelerationField(Interaction):
    def __init__(self, g, name):
        self.g, self.name = g, name

    def set_am(self, system):
        for s in system.sols:
            s.mech_actions[self.name] = MechanicalAction(self.g * s.m, s.get_point(s.g), 0.)


class Spring(Interaction):
    def __init__(self, k, l0, s1, s2, p1, p2):
        self.s1, self.s2, self.p1, self.p2 = s1, s2, p1, p2
        self.l0, self.k = l0, k
        self.name = f'Spring({s2}/{s1})'

    def set_am(self, system):
        a, b = get_point(system, self.s1, self.p1), get_point(system, self.s2, self.p2)
        ab = b - a
        l_ = mag(ab)
        system.sols[self.s1].mech_actions[self.name] = MechanicalAction(ab * self.k * (1 - self.l0 / l_), a, 0.)
        system.sols[self.s2].mech_actions[self.name] = MechanicalAction(ab * self.k * (self.l0 / l_ - 1), b, 0.)


def tmd(system, point, eq) -> np.ndarray:
    return sum((sum(am.babar(point) for am in system.sols[s].mech_actions.values()) for s in eq), np.array((0.,)))


def trd(system, eq) -> np.ndarray:
    return sum((sum(am.f for am in system.sols[s].mech_actions.values()) for s in eq), np.array((0., 0.)))


def p_p_p(system, cycle, rev1, rev2, rev3, _, eq2, eq3):
    (s13, _), (s11, p11) = rev1
    (s21, _), (s22, p22) = rev2
    (s32, _), (s33, p33) = rev3
    p1, p2, p3 = get_point(system, s11, p11), get_point(system, s22, p22), get_point(system, s33, p33)
    p1p2, p3p2 = p2 - p1, p2 - p3

    inv_12, inv_32 = inv_mag(p1p2), inv_mag(p3p2)
    u1, u2 = p1p2 * inv_12, p3p2 * inv_32

    x1, x2 = -tmd(system, p1, eq2 + eq3) * inv_12, -tmd(system, p2, eq2) * inv_32
    y1 = (x2 - dot(u1, u2) * x1) / det(u1, u2)

    f2 = x1 * z_cross(u1) - y1 * u1
    f3 = -trd(system, eq2) - f2
    f1 = -trd(system, eq3) + f3

    n1 = system.joints[cycle[0]].name
    system.sols[s11].mech_actions[n1] = MechanicalAction(-f1, p1, 0.)
    system.sols[s13].mech_actions[n1] = MechanicalAction(f1, p1, 0.)

    n2 = system.joints[cycle[1]].name
    system.sols[s21].mech_actions[n2] = MechanicalAction(-f2, p1, 0.)
    system.sols[s22].mech_actions[n2] = MechanicalAction(f2, p1, 0.)

    n3 = system.joints[cycle[2]].name
    system.sols[s33].mech_actions[n3] = MechanicalAction(-f3, p1, 0.)
    system.sols[s32].mech_actions[n3] = MechanicalAction(f3, p1, 0.)


def p_p_g(system, cycle, rev1, rev2, pri, _, eq2, eq3):
    (s13, _), (s11, p11) = rev1
    (s21, _), (s22, p22) = rev2
    (s1, a1, _), (s1p, _, _) = pri

    p1, p2 = get_point(system, s11, p11), get_point(system, s22, p22)
    u = unit(system.get_ref(s1) + a1)
    n23 = z_cross(u) * (tmd(system, eq3, p1) + (t2 := tmd(system, eq2, p2))) / dot(u, p1 - p2)
    f1 = -trd(system, eq3) - n23
    f2 = -trd(system, eq2) + n23

    n1 = system.joints[cycle[0]].name
    system.sols[s13].mech_actions[n1] = MechanicalAction(f1, p1, 0.)
    system.sols[s11].mech_actions[n1] = MechanicalAction(-f1, p1, 0.)

    n2 = system.joints[cycle[1]].name
    system.sols[s21].mech_actions[n2] = MechanicalAction(-f2, p2, 0.)
    system.sols[s22].mech_actions[n2] = MechanicalAction(f2, p2, 0.)

    n3 = system.joints[cycle[2]].name
    system.sols[s1].mech_actions[n3] = MechanicalAction(-n23, p2, -t2)
    system.sols[s1p].mech_actions[n3] = MechanicalAction(n23, p2, t2)


def g_p_p(system, cycle, pri, rev1, rev2, _, eq2, eq3):
    (s1p, _, _), (s1, a1, _) = pri
    (s11, p11), (s12, _) = rev1
    (s22, p22), (s23, _) = rev2

    p1, p2 = get_point(system, s11, p11), get_point(system, s22, p22)
    u = unit(system.get_ref(s1) + a1)
    n13 = z_cross(u) * ((t3 := tmd(system, eq3, p2)) - tmd(system, eq3 + eq2, p1)) / dot(u, p2 - p1)
    f2 = -trd(system, eq3) - n13
    f1 = -trd(system, eq2) + f2

    n1 = system.joints[cycle[0]].name
    system.sols[s1p].mech_actions[n1] = MechanicalAction(n13, p2, -t3)
    system.sols[s1].mech_actions[n1] = MechanicalAction(-n13, p2, t3)

    n2 = system.joints[cycle[1]].name
    system.sols[s12].mech_actions[n2] = MechanicalAction(f1, p1, 0.)
    system.sols[s11].mech_actions[n2] = MechanicalAction(-f1, p1, 0.)

    n3 = system.joints[cycle[2]].name
    system.sols[s22].mech_actions[n3] = MechanicalAction(-f2, p2, 0)
    system.sols[s23].mech_actions[n3] = MechanicalAction(f2, p2, 0)


def g_g_p(system, cycle, pri1, pri2, rev, _, eq2, eq3):
    (s1p, _, _), (s1, a1, _) = pri1
    (s2, a2, _), (s2p, _, _) = pri2
    (s32, _), (s33, p33) = rev

    p = get_point(system, s33, p33)
    ux, uy = unit(system.get_ref(s1) + a1 + np.pi * .5), unit(system.get_ref(s2) + a2 + np.pi * .5)
    n13, n12 = mat_mul_n(inv_mat(ux, uy), -trd(system, eq3 + eq2))
    n13, n12 = n13 * ux, n12 * uy
    m31, m21 = tmd(system, p, eq3), tmd(system, p, eq2)
    f32 = trd(system, eq3) + n13

    n1 = system.joints[cycle[0]].name
    system.sols[s1p].mech_actions[n1] = MechanicalAction(n13, p, -m31)
    system.sols[s1].mech_actions[n1] = MechanicalAction(-n13, p, m31)

    n2 = system.joints[cycle[1]].name
    system.sols[s2].mech_actions[n2] = MechanicalAction(-n12, p, m21)
    system.sols[s2p].mech_actions[n2] = MechanicalAction(n12, p, -m21)

    n3 = system.joints[cycle[2]].name
    system.sols[s33].mech_actions[n3] = MechanicalAction(-f32, p, 0)
    system.sols[s32].mech_actions[n3] = MechanicalAction(f32, p, 0)


def p_g_g(system, cycle, rev, pri1, pri2, _, eq2, eq3):
    (s13, _), (s11, p11) = rev
    (s1, a1, _), (s1p, _, _) = pri1
    (s2, a2, _), (s2p, _, _) = pri2

    p = get_point(system, s11, p11)
    ux, uy = unit(system.get_ref(s1) + a1 + np.pi * .5), unit(system.get_ref(s2) + a2 + np.pi * .5)
    n21, n23 = mat_mul_n(inv_mat(ux, uy), trd(system, eq2))
    n21, n23 = ux * n21, uy * n23
    f31 = trd(system, eq3) + n23
    m21 = tmd(system, p, eq2 + eq3)
    m23 = tmd(system, p, eq2) - m21

    n1 = system.joints[cycle[0]].name
    system.sols[s11].mech_actions[n1] = MechanicalAction(f31, p, 0)
    system.sols[s13].mech_actions[n1] = MechanicalAction(-f31, p, 0)

    n2 = system.joints[cycle[1]].name
    system.sols[s1].mech_actions[n2] = MechanicalAction(n21, p, m21)
    system.sols[s1p].mech_actions[n2] = MechanicalAction(-n21, p, -m21)

    n3 = system.joints[cycle[2]].name
    system.sols[s2].mech_actions[n3] = MechanicalAction(-n23, p, -m23)
    system.sols[s2p].mech_actions[n3] = MechanicalAction(n23, p, m23)


def sp_p_1(system, cycle, pin, rev, _, eq2):
    (s1, a1, _), (s11, p11) = pin
    (s21, _), (s22, p22) = rev

    p1, p2 = get_point(system, s11, p11), get_point(system, s22, p22)
    u = unit(system.get_ref(s1) + a1)
    n12 = z_cross(u) * tmd(system, p2, eq2) / dot(p2 - p1, u)
    f12 = -trd(system, eq2) - n12

    n1 = system.joints[cycle[0]].name
    system.sols[s11].mech_actions[n1] = MechanicalAction(-n12, p1, 0)
    system.sols[s1].mech_actions[n1] = MechanicalAction(n12, p1, 0)

    n2 = system.joints[cycle[1]].name
    system.sols[s21].mech_actions[n2] = MechanicalAction(-f12, p2, 0)
    system.sols[s22].mech_actions[n2] = MechanicalAction(f12, p2, 0)


def sp_p_2(system, cycle, pin, rev, _, eq2):
    (s12, p12), (s1, a1, _) = pin
    (s21, _), (s22, p22) = rev

    p1, p2 = get_point(system, s12, p12), get_point(system, s22, p22)
    u = unit(system.get_ref(s1) + a1)
    n12 = z_cross(u) * tmd(system, p2, eq2) / dot(p2 - p1, u)
    f12 = -trd(system, eq2) - n12

    n1 = system.joints[cycle[0]].name
    system.sols[s12].mech_actions[n1] = MechanicalAction(n12, p1, 0)
    system.sols[s1].mech_actions[n1] = MechanicalAction(-n12, p1, 0)

    n2 = system.joints[cycle[1]].name
    system.sols[s21].mech_actions[n2] = MechanicalAction(-f12, p2, 0)
    system.sols[s22].mech_actions[n2] = MechanicalAction(f12, p2, 0)


def sp_p(system, cycle, pin, rev, eq1, eq2):
    return (sp_p_1, sp_p_2)[len(pin[0]) == 2](system, cycle, pin, rev, eq1, eq2)


def sp_g_1(system, cycle, pin, pri, _, eq2):
    (s1, a1, _), (s11, p11) = pin
    (s2, a2, _), (s2p, _, _) = pri

    p = get_point(system, s11, p11)
    ux, uy = unit(system.get_ref(s1) + a1 + np.pi * .5), unit(system.get_ref(s2) + a2 + np.pi * .5)
    n21sp, n21g = mat_mul_n(inv_mat(ux, uy), trd(system, eq2))
    n21sp, n21g = n21sp * ux, n21g * uy
    m21 = tmd(system, p, eq2)

    n1 = system.joints[cycle[0]].name
    system.sols[s1].mech_actions[n1] = MechanicalAction(-n21sp, p, 0)
    system.sols[s11].mech_actions[n1] = MechanicalAction(n21sp, p, 0)

    n2 = system.joints[cycle[1]].name
    system.sols[s2].mech_actions[n2] = MechanicalAction(n21g, p, m21)
    system.sols[s2p].mech_actions[n2] = MechanicalAction(-n21g, p, -m21)


def sp_g_2(system, cycle, pin, pri, _, eq2):
    (s12, p12), (s1, a1, _) = pin
    (s2, a2, _), (s2p, _, _) = pri

    p = get_point(system, s12, p12)
    ux, uy = unit(system.get_ref(s1) + a1 + np.pi * .5), unit(system.get_ref(s2) + a2 + np.pi * .5)
    n21sp, n21g = mat_mul_n(inv_mat(ux, uy), trd(system, eq2))
    n21sp, n21g = n21sp * ux, n21g * uy
    m21 = tmd(system, p, eq2)

    n1 = system.joints[cycle[0]].name
    system.sols[s12].mech_actions[n1] = MechanicalAction(-n21sp, p, 0)
    system.sols[s1].mech_actions[n1] = MechanicalAction(n21sp, p, 0)

    n2 = system.joints[cycle[1]].name
    system.sols[s2].mech_actions[n2] = MechanicalAction(n21g, p, m21)
    system.sols[s2p].mech_actions[n2] = MechanicalAction(-n21g, p, -m21)


def sp_g(system, cycle, pin, pri, eq1, eq2):
    (sp_p_1, sp_p_2)[len(pin[0]) == 2](system, cycle, pin, pri, eq1, eq2)


def t_p(system, cycle, rec, rev, _, eq2):
    (s2, _), (s1, _) = rec
    (s11, p11), (s12, _) = rev

    p = get_point(system, s11, p11)
    m21 = tmd(system, p, eq2)
    f21 = trd(system, eq2)

    n1 = system.joints[cycle[0]].name
    system.sols[s1].mech_actions[n1] = MechanicalAction(np.array((0., 0.)), p, m21)
    system.sols[s2].mech_actions[n1] = MechanicalAction(np.array((0., 0.)), p, -m21)

    n2 = system.joints[cycle[1]].name
    system.sols[s11].mech_actions[n2] = MechanicalAction(f21, p, 0)
    system.sols[s12].mech_actions[n2] = MechanicalAction(-f21, p, 0)


def block(system, index, eq1s1, eq2s2):
    system.joints[index].block(system, eq1s1, eq2s2)


dyn = {
    'P_P_P': p_p_p,
    'P_P_G': p_p_g,
    'G_P_P': g_p_p,
    'G_G_P': g_g_p,
    'P_G_G': p_g_g,
    'SP_P': sp_p,
    'SP_G': sp_g,
    'T_P': t_p,
    'Block': block
}



import numpy as np


def rot(theta):  # rotation matrix
    return np.array([[(c := np.cos(theta)), (s := -np.sin(theta))], [-s, c]])


def unit(theta):  # (n,) -> (2, n)
    return np.array([np.cos(theta), np.sin(theta)])


def z_cross(vec):  # (2, ...) -> (2, ...), z ^ v
    return np.array((-vec[1], vec[0]))


def det(v1, v2):  # (2, n) x (2, n) -> (n,)
    return v1[0] * v2[1] - v2[0] * v1[1]


def dot(v1, v2):
    return np.sum(v1 * v2, axis=0)


def mag(vec):  # (2, n) -> (n,)
    return np.sum(vec * vec, axis=0) ** .5


def sq_mag(vec):  # (2, n) -> (n,)
    return np.sum(vec * vec, axis=0)


def inv_mag(vec):  # (2, n) -> (n,)
    return np.sum(vec * vec, axis=0) ** -.5


def inv_mat(u1, u2):  # (2, n) x (2, n) -> (2, 2, n)
    return np.array([[u2[1], -u2[0]], [-u1[1], u1[0]]]) / (u1[0] * u2[1] - u1[1] * u2[0])


def get_angle(vec):  # (2, n) -> (n,)
    return np.arccos(np.maximum(np.minimum(vec[0] * inv_mag(vec), 1.), -1)) * (2 * (vec[1] > 0) - 1)


def get_angle2(vec, im):
    return np.arccos(np.maximum(np.minimum(vec[0] * im, 1.), -1)) * (2 * (vec[1] > 0) - 1)


def mat_mul_n(mat, vec):  # (2, 2, n) x (2, n) -> (2, n), New array
    return np.einsum('ikl,kl->il', mat, vec)


def mat_mul_r(mat, vec):  # (2, 2, n) x (2, n) -> (2, n), Replace array
    np.einsum('ikl,kl->il', mat, vec, out=vec)


def get_point(system, sol, p):
    return system.get_origin(sol) + np.einsum('ikl,k->il', rot(system.get_ref(sol)), p)


def change_ref(system, sol, angle, mat, center, vec):
    # O' = R(angle).(O - center) + vec
    for i in system.eqs[sol]:
        s = system.sols[i]
        s.origin -= center
        mat_mul_r(mat, s.origin)
        s.origin += vec
        s.angle += angle


def change_ref2(system, sol, angle, mat, center):
    # O' = R(angle).(O - center)
    for i in system.eqs[sol]:
        s = system.sols[i]
        s.origin -= center
        mat_mul_r(mat, s.origin)
        s.angle += angle


def trans(system, sol, vec):
    # O' = O + vec
    for i in system.eqs[sol]:
        system.sols[i].origin += vec


# ----------------------------------------------- Calculus -------------------------------------------------------------

def make_continuous(angles):
    l0 = 0
    for i, l in enumerate(angles):
        if not np.isnan(angles[i]):
            k = (l - l0 + np.pi)//(2*np.pi)
            angles[i] -= k*2*np.pi
            l0 = angles[i]
        else:
            l0 = 0


nan = [np.nan]


def derivative1(table, dt):  # f', (n,) -> (n,)
    return np.concatenate((nan, .5 * (table[2:] - table[:-2]) / dt, nan))


def derivative2(table, dt):  # f'', (n,) -> (n,)
    return np.concatenate((nan, (table[2:] + table[:-2] - 2 * table[1:-1]) / (dt * dt), nan))


nan_vec = [np.nan], [np.nan]


def derivative2_vec(table, dt):  # f'', (2, n) -> (2, n)
    return np.concatenate((nan_vec, (table[:, 2:] + table[:, -2] - 2 * table[:, 1:-1]) / (dt * dt), nan_vec), axis=1)



from kinepy.geometry import *
from kinepy.linkage import RevoluteJoint, PinSlotJoint


class Solid:
    def __init__(self, j=0., m=0., g=(0., 0.), name='', rep=0):
        self.rep = rep
        self.name = name if name else f'Solid {self.rep}'
        self.angle = self._points = self.origin = None
        self.j, self.m, self.g = j, m, g

        self.mech_actions = {}

    def get_point(self, point):
        if isinstance(point, (RevoluteJoint, PinSlotJoint)):
            return point.point
        point = np.array(point)
        if point.shape[0] != 2:
            raise ValueError('Shape must start with 2')
        (a00, a01), (a10, a11) = rot(self.angle)
        return self.origin + np.array((point[0] * a00 + point[1] * a01, point[0] * a10 + point[1] * a11))

    def reset(self, n):
        self.mech_actions = {}
        self.origin = np.zeros((2, n), float)
        self.angle = np.zeros((n,), float)

    def save(self):
        return {
            'name': self.name,
            'j': self.j,
            'm': self.m,
            'g': self.g
        }
    
    @classmethod
    def load(cls, data):
        return cls(**data)
    
    def __repr__(self):
        return str(self.rep) + ' | ' + self.name



import numpy as np
import matplotlib.pyplot as plt
from matplotlib import animation
from matplotlib import cm
import matplotlib.colors as mcol

global C
# C = cm.get_cmap('coolwarm') # plasma, coolwarm, bwr, rainbow, OrRd
C = mcol.LinearSegmentedColormap.from_list("", ["cyan", "r"])

def animate(list_paths, list_vectors=None, anim_time=4, repeat=True, scale=1, vector_scale=0.1, magnitude_vector='proportionnal'):
    """
    magnitude_vector : 'proportionnal', 'log', 'unitary'

    """
    if isinstance(list_paths[0], np.ndarray):
        list_paths = [list_paths]
        
    M = []
    x_min, x_max, y_min, y_max = [], [], [], []
    for path in list_paths:
        A = np.array(path)
        nan = np.any(np.isnan(A), axis=(1, 0))
        A[:, :, nan] = 0
        M.append(A.transpose((2, 1, 0)))
        x_min.append(A[:, 0].min())
        x_max.append(A[:, 0].max())
        y_min.append(A[:, 1].min())
        y_max.append(A[:, 1].max())

    x_min, x_max, y_min, y_max = min(x_min), max(x_max), min(y_min), max(y_max)
    dx = (x_max - x_min)*0.05
    dy = (y_max - y_min)*0.05

    fig = plt.figure(figsize=((x_max - x_min)/(y_max - y_min)*scale*7, scale*7))
    ax = plt.axes(xlim=(x_min - dx, x_max + dx), ylim=(y_min - dy, y_max + dy))
    lines = [ax.plot([], [], lw=2, color='tab:blue')[0] for _ in range(len(M))]
    
    if list_vectors is not None:
        if isinstance(list_vectors[0], np.ndarray):
            list_vectors = [list_vectors]
        PAs, vecs = [], []
        for PA, vec in list_vectors:
            PAs.append(PA)
            vecs.append(vec)
        PAs = np.transpose(PAs, (2 ,1, 0))
        vecs = np.transpose(vecs, (2 ,1, 0))
        magnitudes = np.sqrt(np.sum(vecs**2, axis=1)) # linalg
        colors = magnitudes/np.max(magnitudes)
        color_array = np.array([C(color) for color in colors])
        
        if magnitude_vector == 'log':
            a = np.log(magnitudes + 1)/magnitudes
            vecs *= a[:, np.newaxis, :]
        elif magnitude_vector == 'unitary':
            vecs /= magnitudes[:, np.newaxis, :]
        
        n = PAs.shape[2]
        lines.append(ax.quiver([1000]*n,[1000]*n, [0]*n, [0]*n, scale=1/vector_scale, scale_units='xy', cmap='plasma'))
        def _animate_(i):
            for lnum, line in enumerate(lines[:-1]):
                line.set_data(M[lnum][i, 0], M[lnum][i, 1])
            lines[-1].set_UVC(vecs[i, 0], vecs[i, 1])
            lines[-1].set_offsets(PAs[i].T)
            lines[-1].set_facecolor(color_array[i])
            return lines
    else:
        def _animate_(i):
            for lnum, line in enumerate(lines):
                line.set_data(M[lnum][i, 0], M[lnum][i, 1])
            return lines
    
    anim = animation.FuncAnimation(fig, _animate_, frames=(n:=len(M[0])), interval=anim_time*1000/n, blit=True, repeat=repeat)    
    
    return anim

def make_continuous(L):
    l0 = 0
    for i, l in enumerate(L):
        if not np.isnan(L[i]):
            k = (l - l0 + np.pi)//(2*np.pi)
            L[i] -= k*2*np.pi
            l0 = L[i]
    return L

distance = lambda p1, p2: np.sqrt(np.sum((p2 - p1)**2, axis=0))

to_cartesian = lambda r, a : (r*np.cos(a), r*np.sin(a))

def get_speed(L, t):
    if len(L.shape) == 1:    
        return np.diff(L)/t*(len(L) - 1)
    return np.diff(L, axis=1)/t*(L.shape[1] - 1)


def get_speed_accurate(L, t):
    if len(L.shape) == 1:    
        return (L[2:] - L[:-2])/(2*t/(len(L) - 1))
    return (L[:, 2:] - L[:, :-2])/(2*t/(L.shape[1] - 1))


def get_acceleration(L, t):
    if len(L.shape) == 1:    
        return np.diff(L, 2)/(t/(len(L) - 1))**2
    return np.diff(L, 2, axis=1)/(t/(L.shape[1] - 1))**2



from setuptools import setup

setup(
    name='kinepy',
    packages=['kinepy'],
    version='0.0.1',
    description='Python library dedicated to plane mechanism simulation',
    author='Loïc Chevalier - Valentin Burillier',
    url='https://github.com/valentin-burillier/kinepy',
    keywords=['robotics', 'dynamic', 'kinematics', 'static', 'inverse-kinematics', 'mechanism'],
    requires=[],
)
